import os
import warnings

import autonbox
import cv2
import numpy as np
import torch
import typing
from autonbox.contrib.resnet.dataset import Video
from autonbox.contrib.resnet.model import generate_model
from autonbox.contrib.resnet.spatial_transforms import (Compose, Normalize, Scale, CenterCrop, ToTensor)
from autonbox.contrib.resnet.temporal_transforms import LoopPadding
from d3m import container, utils as d3m_utils
from d3m.metadata import base as metadata_base, hyperparams
from d3m.primitive_interfaces import base
from d3m.primitive_interfaces.base import CallResult
from d3m.primitive_interfaces.featurization import FeaturizationTransformerPrimitiveBase
from torch.autograd import Variable

Inputs = container.DataFrame
Outputs = container.DataFrame


class Hyperparams(hyperparams.Hyperparams):
    num_workers = hyperparams.Hyperparameter[int](
        semantic_types=['https://metadata.datadrivendiscovery.org/types/ResourcesUseParameter'],
        default=0,
        description='The number of subprocesses to use for data loading. 0 means that the data will be loaded in the '
                    'main process.'
    )


class ResNext101KineticsParams(object):
    # TODO determine which to make as hyper-parameters
    n_classes = 400
    mode = 'feature'
    clip_vid = 'mean'
    down_rate = 1
    model_name = 'resnext'
    model_depth = 101
    resnet_shortcut = 'B'
    resnext_cardinality = 32
    sample_size = 112
    sample_duration = 16  # number of frames in one clip
    no_cuda = True
    mean = [114.7748, 107.7354, 99.4750]
    batch_size = 32


class ResNext101KineticsPrimitive(FeaturizationTransformerPrimitiveBase[Inputs, Outputs, Hyperparams]):
    """
    Video Feature Extraction for Action Classification With 3D ResNet
    """
    metadata = metadata_base.PrimitiveMetadata(
        {
            'id': '15935e70-0605-4ded-87cf-2933ca35d4dc',
            'version': '0.2.0',
            "name": "Video Feature Extraction for Action Classification With 3D ResNet",
            'description': "Video Feature Extraction for Action Classification With 3D ResNet",
            'python_path': 'd3m.primitives.feature_extraction.resnext101_kinetics_video_features.VideoFeaturizer',
            'source': {
                'name': autonbox.__author__,
                'uris': ['https://github.com/autonlab/autonbox'],
                'contact': 'mailto:awillia2@andrew.cmu.edu'
            },
            'installation': [{
                'type': metadata_base.PrimitiveInstallationType.PIP,
                'package': 'autonbox',
                'version': '0.2.4'
            }, {
                'type': metadata_base.PrimitiveInstallationType.FILE,
                'key': autonbox.__key_static_file_resnext__,
                'file_uri': 'http://public.datadrivendiscovery.org/resnext-101-kinetics.pth',
                'file_digest': autonbox.__digest_static_file_resnext__
            }],
            'algorithm_types': [
                metadata_base.PrimitiveAlgorithmType.CONVOLUTIONAL_NEURAL_NETWORK,
            ],
            'primitive_family': metadata_base.PrimitiveFamily.FEATURE_EXTRACTION,
        },
    )

    def __init__(self, *, hyperparams: Hyperparams, volumes: typing.Union[typing.Dict[str, str], None] = None) -> None:
        super().__init__(hyperparams=hyperparams, volumes=volumes)

        self._config = ResNext101KineticsParams

        torch.manual_seed(self.random_seed)  # seed the RNG for all devices (both CPU and CUDA):

        # Use GPU if available
        if torch.cuda.is_available():
            self.logger.info("Use GPU.")
            self._config.no_cuda = False
            # For reproducibility on CuDNN backend
            torch.backends.cudnn.deterministic = True
            torch.backends.cudnn.benchmark = False
        else:
            self.logger.info("Use CPU.")
            self._config.no_cuda = True

        self.logger.info('Number of workers: {}'.format(self.hyperparams['num_workers']))

        self._down_rate = 1

    def _instantiate_model(self):
        model = generate_model(self._config)
        model_data = self._load_model()
        if self._config.no_cuda:
            state_dict = {k.replace('module.', ''): v for k, v in model_data['state_dict'].items()}  # for cpu only
        else:
            state_dict = model_data['state_dict']
        model.load_state_dict(state_dict)
        model.eval()
        return model

    def produce(self, *, inputs: Inputs, timeout: float = None, iterations: int = None) -> CallResult[Outputs]:
        """
        :param inputs: assume the first column is the filename
        :param timeout:
        :param iterations:
        :return:
        """
        model = self._instantiate_model()
        features = []
        # TODO consider a more robust means to 1) get location_base_uris and remove file://
        media_root_dir = inputs.metadata.query((0, 0))['location_base_uris'][0][len('file://'):]  # remove file://
        for filename in inputs.iloc[:, 0]:
            file_path = os.path.join(media_root_dir, filename)
            if os.path.isfile(file_path):
                video = self._read_fileuri(file_path)  # video is a ndarray of F x H x W x C, e.g. (408, 240, 320, 3)
                feature = self._generate_vid_feature(model, video)
            else:
                self.logger.warning("No such file {}. Feature vector will be set to all zeros.".format(file_path))
                feature = np.zeros(2048)
            features.append(feature)

        results = container.DataFrame(features, generate_metadata=True)

        return base.CallResult(results)

    def _generate_vid_feature(self, model, vid_matrix):
        """
        Modified from function classify_video()
        :param vid_matrix: takes in video matrix F(frames) x H(height) x W(width) x C(channels)
                           dtype of matrix is uint8
        :return: ndarray representation of video
        """
        assert vid_matrix.ndim == 4 and self._down_rate <= 1  # sanity check
        spatial_transform = Compose([Scale(self._config.sample_size),
                                     CenterCrop(self._config.sample_size),
                                     ToTensor(),
                                     Normalize(self._config.mean, [1, 1, 1])])
        temporal_transform = LoopPadding(self._config.sample_duration)
        data = Video(vid_matrix, spatial_transform=spatial_transform,
                     temporal_transform=temporal_transform,
                     sample_duration=self._config.sample_duration, down_rate=self._down_rate)
        data_loader = torch.utils.data.DataLoader(data, batch_size=self._config.batch_size,
                                                  num_workers=self.hyperparams['num_workers'],
                                                  shuffle=False, pin_memory=True)
        video_outputs = []
        with torch.no_grad():
            for i, inputs in enumerate(data_loader):
                inputs = Variable(inputs)
                # input is of shape n x 3 x sample_duration x 112 x 112
                outputs = model(inputs)
                # output is of format n(batch size) x d(dimension of feature)
                video_outputs.append(outputs.cpu().data)
        video_outputs = np.concatenate(video_outputs, axis=0)
        mean_feature = np.mean(video_outputs, axis=0)  # shape of (d, )
        return mean_feature

    def _load_model(self):
        """
        Loads the model from the volume
        :return:
        """
        key_filename = autonbox.__key_static_file_resnext__
        static_dir = os.getenv('D3MSTATICDIR', '/static')
        if key_filename in self.volumes:
            _weight_file_path = self.volumes[key_filename]
            self.logger.info("Weights file path found in static volumes")
        else:
            self.logger.info("Trying to locate weights file in the static folder {}".format(static_dir))
            _weight_file_path = os.path.join(static_dir, autonbox.__digest_static_file_resnext__)

        if os.path.isfile(_weight_file_path):
            if torch.cuda.is_available():  # GPU
                model_data = torch.load(_weight_file_path)
            else:  # CPU only
                model_data = torch.load(_weight_file_path, map_location='cpu')
            self.logger.info("Loaded weights file")
        else:
            raise ValueError("Can't get weights file from the volume by key: {} or in the static folder: {}".format(
                key_filename, static_dir))

        return model_data

    def _read_fileuri(self, fileuri: str) -> container.ndarray:
        """
        @see https://gitlab.com/datadrivendiscovery/common-primitives/blob/master/common_primitives/video_reader.py#L65
        :param fileuri:
        :return:
        """
        capture = cv2.VideoCapture(fileuri)
        frames = []

        try:
            while capture.isOpened():
                ret, frame = capture.read()
                if not ret:
                    break
                else:
                    assert frame.dtype == np.uint8, frame.dtype

                    if frame.ndim == 2:
                        # Make sure there are always three dimensions.
                        frame = frame.reshape(list(frame.shape) + [1])

                    assert frame.ndim == 3, frame.ndim

                    frames.append(frame)
        finally:
            capture.release()

        return container.ndarray(np.array(frames), generate_metadata=False)



import os
import typing
import numpy as np
import pandas as pd

from d3m import container
from d3m import utils as d3m_utils
from d3m.metadata import base as metadata_base, hyperparams
from d3m.primitive_interfaces.base import CallResult
from d3m.primitive_interfaces.transformer import TransformerPrimitiveBase

import autonbox

Inputs = container.List
Outputs = container.DataFrame

class MergePartialPredictionsPrimitive(TransformerPrimitiveBase[Inputs, Outputs, hyperparams.Hyperparams]):
    """
        Merge predictions of multiple models
        Useful if model do not produce predictions for each points and that it is necessary
        to merge those predictions (the first non nan will be returned)
    """

    metadata = metadata_base.PrimitiveMetadata({
        'id': '1cc95f70-0716-11ea-9762-3dd2bb86dde8',
        'version': '0.1.0',
        'name': "Merge predictions of multiple models",
        'python_path': 'd3m.primitives.data_transformation.merge_partial_predictions.AutonBox',
        'source': {
            'name': autonbox.__author__,
            'uris': ['https://github.com/autonlab/autonbox'],
            'contact': 'mailto:vjeansel@andrew.cmu.edu'
        },
        "installation": [{
            "type": metadata_base.PrimitiveInstallationType.PIP,
            "package": "autonbox",
            "version": "0.2.4"
        }],
        'algorithm_types': [
            metadata_base.PrimitiveAlgorithmType.DATA_CONVERSION, #TODO: Choose
        ],
        'primitive_family': metadata_base.PrimitiveFamily.DATA_TRANSFORMATION
    })

    def produce(self, *, inputs: Inputs, timeout: float = None, iterations: int = None) -> CallResult[Outputs]:
        # Force index to be d3mIndex for concatenation
        inputs_reindex = []
        for i in inputs:
            i_copy = i.copy()
            i_copy.index = i_copy.d3mIndex
            i_copy = i_copy.drop(columns="d3mIndex")
            inputs_reindex.append(i_copy)
        
        # Merge inputs       
        output = pd.concat(inputs_reindex, axis = 1)
        output.metadata = inputs[-1].metadata
        
        # Propagate best non nan score
        output = output.T.fillna(method = 'bfill').T.iloc[:, :1]
        output = output.reset_index().dropna()

        return CallResult(output)



import os
import pandas as pd

from d3m import container
from d3m import utils as d3m_utils
from d3m.metadata import base as metadata_base, hyperparams
from d3m.primitive_interfaces.base import CallResult
from d3m.primitive_interfaces.transformer import TransformerPrimitiveBase

import autonbox

Inputs = container.Dataset
Outputs = container.Dataset

class Hyperparams(hyperparams.Hyperparams):
    original_columns = hyperparams.Set(
        elements=hyperparams.Hyperparameter[str](''),
        default=(),
        semantic_types=['https://metadata.datadrivendiscovery.org/types/ControlParameter'],
        description="List of the original columns"
    )
    percentage_missing = hyperparams.Uniform(
        lower=0.0, 
        upper=1.0, 
        default=0.5,
        semantic_types=['https://metadata.datadrivendiscovery.org/types/TuningParameter'],
        description='Percentage of missing data allowed (above this percentage and the line is deleted)'
    )


class CleanAugmentationPrimitive(TransformerPrimitiveBase[Inputs, Outputs, Hyperparams]):
    """
        Remove rows which haven't been augmented following the datamart augmentation 
        (any rows with more than percentage_missing columns which are resulting from augmentation)
        
        NB: This primitive results might reduce the number of row
    """

    metadata = metadata_base.PrimitiveMetadata({
        'id': 'fe0f1ac8-1d39-463a-b344-7bd498a31b92',
        'version': '0.1.0',
        'name': "Clean dataset of unaugmented rows",
        'python_path': 'd3m.primitives.data_cleaning.clean_augmentation.AutonBox',
        'source': {
            'name': autonbox.__author__,
            'uris': ['https://github.com/autonlab/autonbox'],
            'contact': 'mailto:vjeansel@andrew.cmu.edu'
        },
        "installation": [{
            "type": metadata_base.PrimitiveInstallationType.PIP,
            "package": "autonbox",
            "version": "0.2.4"
        }],
        'algorithm_types': [
            #metadata_base.PrimitiveAlgorithmType.ROW_SELECTION
            metadata_base.PrimitiveAlgorithmType.DATA_RETRIEVAL, #TODO: Delete when new algo released
        ],
        'primitive_family': metadata_base.PrimitiveFamily.DATA_CLEANING
    })

    def produce(self, *, inputs: Inputs, timeout: float = None, iterations: int = None) -> CallResult[Outputs]:
        # Selection augmentation columns
        augmented_data = inputs['learningData'][[c for c in inputs['learningData'].columns if c not in self.hyperparams['original_columns']]]
        
        # Count absent data: at Dataset level: empty field
        absent = (augmented_data == '').mean(axis = 1)
        
        # Selection of the augmented lines
        output = inputs.copy()
        output['learningData'] = output['learningData'].loc[absent < self.hyperparams['percentage_missing']]

        return CallResult(output)



import torch
import torch.nn as nn
import torch.nn.functional as F
from collections import OrderedDict
import math

__all__ = ['DenseNet', 'densenet121', 'densenet169', 'densenet201', 'densenet264']


def densenet121(**kwargs):
    model = DenseNet(num_init_features=64, growth_rate=32, block_config=(6, 12, 24, 16),
                     **kwargs)
    return model


def densenet169(**kwargs):
    model = DenseNet(num_init_features=64, growth_rate=32, block_config=(6, 12, 32, 32),
                     **kwargs)
    return model


def densenet201(**kwargs):
    model = DenseNet(num_init_features=64, growth_rate=32, block_config=(6, 12, 48, 32),
                     **kwargs)
    return model


def densenet264(**kwargs):
    model = DenseNet(num_init_features=64, growth_rate=32, block_config=(6, 12, 64, 48),
                     **kwargs)
    return model


def get_fine_tuning_parameters(model, ft_begin_index):
    if ft_begin_index == 0:
        return model.parameters()

    ft_module_names = []
    for i in range(ft_begin_index, 5):
        ft_module_names.append('denseblock{}'.format(ft_begin_index))
        ft_module_names.append('transition{}'.format(ft_begin_index))
    ft_module_names.append('norm5')
    ft_module_names.append('classifier')

    parameters = []
    for k, v in model.named_parameters():
        for ft_module in ft_module_names:
            if ft_module in k:
                parameters.append({'params': v})
                break
        else:
            parameters.append({'params': v, 'lr': 0.0})

    return parameters


class _DenseLayer(nn.Sequential):
    def __init__(self, num_input_features, growth_rate, bn_size, drop_rate):
        super(_DenseLayer, self).__init__()
        self.add_module('norm.1', nn.BatchNorm3d(num_input_features))
        self.add_module('relu.1', nn.ReLU(inplace=True))
        self.add_module('conv.1', nn.Conv3d(num_input_features, bn_size * growth_rate,
                                            kernel_size=1, stride=1, bias=False))
        self.add_module('norm.2', nn.BatchNorm3d(bn_size * growth_rate))
        self.add_module('relu.2', nn.ReLU(inplace=True))
        self.add_module('conv.2', nn.Conv3d(bn_size * growth_rate, growth_rate,
                                            kernel_size=3, stride=1, padding=1, bias=False))
        self.drop_rate = drop_rate

    def forward(self, x):
        new_features = super(_DenseLayer, self).forward(x)
        if self.drop_rate > 0:
            new_features = F.dropout(new_features, p=self.drop_rate, training=self.training)
        return torch.cat([x, new_features], 1)


class _DenseBlock(nn.Sequential):
    def __init__(self, num_layers, num_input_features, bn_size, growth_rate, drop_rate):
        super(_DenseBlock, self).__init__()
        for i in range(num_layers):
            layer = _DenseLayer(num_input_features + i * growth_rate, growth_rate, bn_size, drop_rate)
            self.add_module('denselayer%d' % (i + 1), layer)


class _Transition(nn.Sequential):
    def __init__(self, num_input_features, num_output_features):
        super(_Transition, self).__init__()
        self.add_module('norm', nn.BatchNorm3d(num_input_features))
        self.add_module('relu', nn.ReLU(inplace=True))
        self.add_module('conv', nn.Conv3d(num_input_features, num_output_features,
                                          kernel_size=1, stride=1, bias=False))
        self.add_module('pool', nn.AvgPool3d(kernel_size=2, stride=2))


class DenseNet(nn.Module):
    """Densenet-BC model class
    Args:
        growth_rate (int) - how many filters to add each layer (k in paper)
        block_config (list of 4 ints) - how many layers in each pooling block
        num_init_features (int) - the number of filters to learn in the first convolution layer
        bn_size (int) - multiplicative factor for number of bottle neck layers
          (i.e. bn_size * k features in the bottleneck layer)
        drop_rate (float) - dropout rate after each dense layer
        num_classes (int) - number of classification classes
    """
    def __init__(self, sample_size, sample_duration, growth_rate=32, block_config=(6, 12, 24, 16),
                 num_init_features=64, bn_size=4, drop_rate=0, num_classes=1000, last_fc=True):

        super(DenseNet, self).__init__()

        self.last_fc = last_fc

        self.sample_size = sample_size
        self.sample_duration = sample_duration

        # First convolution
        self.features = nn.Sequential(OrderedDict([
            ('conv0', nn.Conv3d(3, num_init_features, kernel_size=7,
                                stride=(1, 2, 2), padding=(3, 3, 3), bias=False)),
            ('norm0', nn.BatchNorm3d(num_init_features)),
            ('relu0', nn.ReLU(inplace=True)),
            ('pool0', nn.MaxPool3d(kernel_size=3, stride=2, padding=1)),
        ]))

        # Each denseblock
        num_features = num_init_features
        for i, num_layers in enumerate(block_config):
            block = _DenseBlock(num_layers=num_layers, num_input_features=num_features,
                                bn_size=bn_size, growth_rate=growth_rate, drop_rate=drop_rate)
            self.features.add_module('denseblock%d' % (i + 1), block)
            num_features = num_features + num_layers * growth_rate
            if i != len(block_config) - 1:
                trans = _Transition(num_input_features=num_features, num_output_features=num_features // 2)
                self.features.add_module('transition%d' % (i + 1), trans)
                num_features = num_features // 2

        # Final batch norm
        self.features.add_module('norm5', nn.BatchNorm2d(num_features))

        # Linear layer
        self.classifier = nn.Linear(num_features, num_classes)

    def forward(self, x):
        features = self.features(x)
        out = F.relu(features, inplace=True)
        last_duration = math.ceil(self.sample_duration / 16)
        last_size = math.floor(self.sample_size / 32)
        out = F.avg_pool3d(out, kernel_size=(last_duration, last_size, last_size)).view(features.size(0), -1)
        if self.last_fc:
            out = self.classifier(out)
        return out



import random
import math
import numbers
import collections
import numpy as np
import torch
from PIL import Image, ImageOps
try:
    import accimage
except ImportError:
    accimage = None


class Compose(object):
    """Composes several transforms together.
    Args:
        transforms (list of ``Transform`` objects): list of transforms to compose.
    Example:
        >>> transforms.Compose([
        >>>     transforms.CenterCrop(10),
        >>>     transforms.ToTensor(),
        >>> ])
    """

    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, img):
        for t in self.transforms:
            img = t(img)
        return img


class ToTensor(object):
    """Convert a ``PIL.Image`` or ``numpy.ndarray`` to tensor.
    Converts a PIL.Image or numpy.ndarray (H x W x C) in the range
    [0, 255] to a torch.FloatTensor of shape (C x H x W) in the range [0.0, 1.0].
    """

    def __call__(self, pic):
        """
        Args:
            pic (PIL.Image or numpy.ndarray): Image to be converted to tensor.
        Returns:
            Tensor: Converted image.
        """
        if isinstance(pic, np.ndarray):
            # handle numpy array
            img = torch.from_numpy(pic.transpose((2, 1, 0)))
            # backward compatibility
            return img.float()

        if accimage is not None and isinstance(pic, accimage.Image):
            nppic = np.zeros([pic.channels, pic.height, pic.width], dtype=np.float32)
            pic.copyto(nppic)
            return torch.from_numpy(nppic)

        # handle PIL Image
        if pic.mode == 'I':
            img = torch.from_numpy(np.array(pic, np.int32, copy=False))
        elif pic.mode == 'I;16':
            img = torch.from_numpy(np.array(pic, np.int16, copy=False))
        else:
            img = torch.ByteTensor(torch.ByteStorage.from_buffer(pic.tobytes()))
        # PIL image mode: 1, L, P, I, F, RGB, YCbCr, RGBA, CMYK
        if pic.mode == 'YCbCr':
            nchannel = 3
        elif pic.mode == 'I;16':
            nchannel = 1
        else:
            nchannel = len(pic.mode)
        img = img.view(pic.size[1], pic.size[0], nchannel)
        # put it from HWC to CHW format
        # yikes, this transpose takes 80% of the loading time/CPU
        img = img.transpose(0, 1).transpose(0, 2).contiguous()
        if isinstance(img, torch.ByteTensor):
            return img.float()
        else:
            return img


class Normalize(object):
    """Normalize an tensor image with mean and standard deviation.
    Given mean: (R, G, B) and std: (R, G, B),
    will normalize each channel of the torch.*Tensor, i.e.
    channel = (channel - mean) / std
    Args:
        mean (sequence): Sequence of means for R, G, B channels respecitvely.
        std (sequence): Sequence of standard deviations for R, G, B channels
            respecitvely.
    """

    def __init__(self, mean, std):
        self.mean = mean
        self.std = std

    def __call__(self, tensor):
        """
        Args:
            tensor (Tensor): Tensor image of size (C, H, W) to be normalized.
        Returns:
            Tensor: Normalized image.
        """
        # TODO: make efficient
        for t, m, s in zip(tensor, self.mean, self.std):
            t.sub_(m).div_(s)
        return tensor


class Scale(object):
    """Rescale the input PIL.Image to the given size.
    Args:
        size (sequence or int): Desired output size. If size is a sequence like
            (w, h), output size will be matched to this. If size is an int,
            smaller edge of the image will be matched to this number.
            i.e, if height > width, then image will be rescaled to
            (size * height / width, size)
        interpolation (int, optional): Desired interpolation. Default is
            ``PIL.Image.BILINEAR``
    """

    def __init__(self, size, interpolation=Image.BILINEAR):
        assert isinstance(size, int) or (isinstance(size, collections.Iterable) and len(size) == 2)
        self.size = size
        self.interpolation = interpolation

    def __call__(self, img):
        """
        Args:
            img (PIL.Image): Image to be scaled.
        Returns:
            PIL.Image: Rescaled image.
        """
        if isinstance(self.size, int):
            if isinstance(img, np.ndarray):
                [h,w,c] = img.shape
                if (w <= h and w == self.size) or (h <= w and h == self.size):
                    return img
                if w < h:
                    ow = self.size
                    oh = int(self.size * h / w)
                    return np.resize(img,[oh, ow, c])
                else:
                    oh = self.size
                    ow = int(self.size * w / h)
                    return np.resize(img,[oh, ow, c])
            w, h = img.size
            if (w <= h and w == self.size) or (h <= w and h == self.size):
                return img
            if w < h:
                ow = self.size
                oh = int(self.size * h / w)
                return img.resize((ow, oh), self.interpolation)
            else:
                oh = self.size
                ow = int(self.size * w / h)
                return img.resize((ow, oh), self.interpolation)
        else:
            return img.resize(self.size, self.interpolation)


class CenterCrop(object):
    """Crops the given PIL.Image at the center.
    Args:
        size (sequence or int): Desired output size of the crop. If size is an
            int instead of sequence like (h, w), a square crop (size, size) is
            made.
    """

    def __init__(self, size):
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size

    def __call__(self, img):
        """
        Args:
            img (PIL.Image): Image to be cropped.
        Returns:
            PIL.Image: Cropped image.
        """
        if isinstance(img, np.ndarray):
            # handle numpy array
            [h,w] = img.shape[:2]
            th, tw = self.size
            x1 = int(round((w - tw) / 2.))
            y1 = int(round((h - th) / 2.))
            # print(img[x1:x1+tw, y1:y1+th,:].shape)
            return img[y1:y1+th,x1:x1+tw,:]
        else:
            w, h = img.size
            th, tw = self.size
            x1 = int(round((w - tw) / 2.))
            y1 = int(round((h - th) / 2.))
            return img.crop((x1, y1, x1 + tw, y1 + th))
